#!/usr/bin/perl
# qdp
#
# Authored by Chris Gray (fergus@azathoth.com)
#
#-------------------------------------------------------------------------------
#
# Description:
#
# A 'Quick And Dirty' preforking, multithreaded HTTP proxy server which provides
# several security-oriented functions.
#
#-------------------------------------------------------------------------------
#
# Subroutines:
#
#	- daemonify
#		Turns the process into a daemon by redirecting the STD handles,
#		changes the working directory to "/", sets the umask to 0, and
#		forks itself into the background, thereby disassociating itself
#		from the calling terminal.
#
# 	- acceptConnections
#		Attempts to set up the server on the port specified by $PORT and
#		if successful, enters a loop which checks for a connection on
#		$PORT and then spawns a child process which deals with the
#		actual data, the main process waits for another connection.
#
# 	- handleConnection
#		Like the name implies, it contains (most) of the connection
#		management code, with the exception of response retrieval, which
#		is handled by 'getRequest'.
#
# 	- response getRequest
#		Attempts to open a connection with the requested server and if
#		successful, sends the request and gives the 'response' back to
#		'handleConnection'.
#
# 	- bool allowedDeniedSite(current_requested_site,mode)
#		Checks the currently requested site against all of the allowed
#		sites (contained in @ALLOWED), or all of the denied sites
#		(contained in @DENIED), based upon 'mode' and returns a boolean
#		based upon the outcome. '0' for NOT OK, and '1' for OK.
#
# 	- site_list getAllowedDeniedSites(path)
#		Reads in the list of specified hosts hosts, contained within the
#		 'hosts.allow' or 'host.deny' files and returns them as
#		site_list. 
#
# 	- debug
#		Turns on transaction logging.
#
# 	- reinit
#		Closes and reinitializes all connections and refreshes the
#		allowed and denied sites lists.
#
# 	- shutDown
#		Cleanly closes all open connections and then exits.
#
# 	- filtered_data filter(raw_data)
#		Runs raw_data through each active HTML filter, and returns it as
#		filtered_data.
#
# 	- filtered_data selectFilter(raw_data)
#		Adds a 'size="4"' to each select tag in raw_data, thereby
#		transforming all drop-down menus into scrolling list boxes.
#
# 	- filtered_data windowFilter(raw_data)
#		Breaks every 'window.open' command in raw_data.
#
# 	- filtered_data targetFilter(raw_data)
#		Breaks certain 'target' tags in raw_data.
#
# 	- filtered_data templateFilter(raw_data)
#		Provided as a template for later filters.
#
# 	- version getVers
#		Gets the current linux version from /proc/version.
#
# 	- verifyCommandLine
#		Checks for the validity of the command line parameters and
#		exits the program with an error, or changes the entries in the
#		%ARG hash to something other than the default.
#
# 	- forbidden
#		Error handling for disallowed site requests.
#
# 	- unavailable
#		Error handling for a 'down network' condition.
#

use IO::Socket;

# Command line argument defaults;
%ARGS=(
	'-browser'	=> 0,			   # Change browser type
	'-denyfile'	=> "/etc/qdp/hosts.deny",  # Default path
	'-filtering'	=> 1,			   # Initially ON
	'-hostfile'	=> "/etc/qdp/hosts.allow", # Default path
	'-log'		=> 0,			   # Initially OFF
	'-maxconn'	=> SOMAXCONN,		   # Module Default
	'-port'		=> 8080,		   # STD HTTP proxy port
	'-sitelock'	=> 0,			   # Initially OFF
	'-timeout'	=> 15,			   # 15 seconds default
      );

# Change the default arguments if specified
if(length(@ARGV[0])){verifyCommandLine();}

# Constants
$EOL       = "\015\012";
$BLANK     = $EOL x 2;

$BROWSER   = $ARGS{'-browser'};
$DENYPATH  = $ARGS{'-denyfile'};
$FILTERING = $ARGS{'-filtering'};
$CONFPATH  = $ARGS{'-hostfile'};
$LOG       = $ARGS{'-log'};
$MAXCONN   = $ARGS{'-maxconn'};
$PORT      = $ARGS{'-port'};
$SITELOCK  = $ARGS{'-sitelock'};
$TIMEOUT   = $ARGS{'-timeout'};

$USERAGENT = "monkeyzilla/0.1 [en] (Linux ".getVers().")";
@ALLOWED   = getAllowedDeniedSites($CONFPATH);
@DENIED    = getAllowedDeniedSites($DENYPATH);

# Filter subroutines (constant)
@FILTERS=(
	  \&selectFilter,
	  \&windowFilter,
	  \&targetFilter
	 );

# Declare socket handles
$REMOTE=undef;
$CLIENT=undef;
$SERVER=undef;

# Assign routines to certain signals
$SIG{'USR1'}=\&shutDown;
$SIG{'USR2'}=\&debug;
$SIG{'PIPE'}=\&reinit;

# Daemonize the process
daemonify();

# Begin waiting for requests
acceptConnections();


################################ daemonization #################################


sub daemonify
{
	# Redirect the STD file handles
	open(STDIN,"/dev/null");
	open(STDOUT,">>/tmp/proxy.log");
	open(STDERR,">>/tmp/proxy.err");

	# Change the working directory to root
	chdir("/");

	# Set umask to 0
	umask(0);

	# Fork a daemon process
	fork && exit;
}


######################## connection handling functions #########################


sub acceptConnections
{
	# Set up the server socket
	$SERVER=IO::Socket::INET->new(
				      Proto     => 'tcp',
				      LocalPort => $PORT,
				      Listen    => $MAXCONN,
				      Reuse     => 1
				     );

	# Error out if not able to connect to chosen port
	unless($SERVER)
	{
		print STDERR time(),"\nCan't set up server on port: $PORT\n\n";
		shutDown();
	}

	# Sit indefinitely, waiting for client connections.
	for(;($CLIENT=$SERVER->accept());close($CLIENT))
	{
		# Fork (using the double-fork trick described 
		# in the perlfunc section of the standard perl
		# documentation) into a child process to complete
		# order and free up the parent to take another request.
		unless($pid=fork)
		{
			unless(fork)
			{
	 			handleConnection();
				exit();
			}
	 		exit();
	 	}

		waitpid($pid,0);
	}
}

sub handleConnection
{
	# Initialize counter, booleans, and buffer to default values
	$emptyLineCnt=0;	# How many empty lines we have
	$checkForPost=0;	# Boolean that lets the POST/GET
				# check happen only once.
	$isPost=0;		# POST/GET boolean
	$request='';		# Request buffer

	# As long as the client is sending data, receive it one line at a time.
	while(<$CLIENT>)
	{
		# Append current line to the buffer
		$request.= $_;

		# First time through, check to see if the
		# request is of type POST, or of type GET.
		if($checkForPost == 0)
		{
			if(substr($request,0,4) eq "POST")
			{
				$isPost=1;
			}

			$checkForPost=1;
		}

		# If the request is in the form of a POST
		# there are two end-of-line terminators
		# due to the presence of POSTed data.
		if($isPost == 1)
		{
			if($emptyLineCnt == 1)
			{
				$request.=$EOL;
				last;
			}

			if($_ eq $EOL)
			{
				$emptyLineCnt++;
			}
		}
		# If the request is in the form of a GET
		# then there should only be one end-of-line
		# terminator.
		else
		{
			if($_ eq $EOL)
			{
				$request.=$EOL;
				last;
			}
		}
	}

	# Parse out the server's address.
	if($request=~/\b(Host:)\s+(\S+)/i){$address=$2;}

	# Make it so only the document path is part of the request
	# I.E. remove the leading 'http(s)://mydomain.com'.
	$request=~s{http(|s)://.*?/}{/}i;

	# Replace the user agent with our own custom one.
	if($BROWSER && !allowedDeniedSite($address,0))
	{
		if($request=~/\b(User-Agent:)\s+(.*)/i){$userAgent=$2;}
		$request=~s/\Q$userAgent\E/$USERAGENT/;
	}

	# Try to get the file we are requesting. Wrap a connection
	# error handler around it JIC.
	unless($response=getRequest($address,$request))
	{
		$response=unavailable($address);
	}

	# Split a copy of the response header for the archives...
	($rspHeader,undef)=split(/$BLANK/,$response);

	# If incoming data is of type 'text' then run it through
	# our filter(s).
	if($rspHeader=~/\b(Content-Type:)\s+(\w+)/i){$cType=lc($2);}

	if(($FILTERING == 1) && ($cType eq "text"))
	{
		$response=filter($response);
	}

	# If the address requested is not in our hosts.allow file, then
	# Redirect them to an error page. Note: This should only happen
	# when the item requested is a document of type 'text'... Can go
	# awry if code from an unallowed site is referenced by an acceptable
	# site.
	if(
	   ($SITELOCK == 1) && 
	   !allowedDeniedSite($address,1) && 
	   ($cType eq "text")
	  )
	{
		$response=forbidden($address);
	}

	# If logging is enabled, then print log data.
	if($LOG)
	{
		print STDOUT "\n",time(),"\n$request\n----\n$rspHeader\n####\n";
	}

	# Give the client its requested content
	print $CLIENT $response;
}

sub getRequest()
{
	my $return;

	# Create inet socket for our remote connection
	$REMOTE=IO::Socket::INET->new(
					Proto	 => "tcp",
					PeerAddr => @_[0],
					PeerPort => "http(80)",
					Timeout  => $TIMEOUT
                                     );

	# The connection was accepted
	if($REMOTE)
	{
		# Send request to server
		print $REMOTE @_[1];

		# Retrieve request from server
		while(<$REMOTE>){$return.=$_;}

		# Terminate connection to server
		close $REMOTE;
	}
	# The connection was not accepted
	else
	{
		$return=unavailable(@_[0]);
	}

	return($return);
}


########################## site restriction functions ##########################


sub allowedDeniedSite($$)
{
	# Run through each entry in the allowed
	# sites list and check each entry against
	# the currently requested site.
	if(@_[1])
	{
		for(my $i=0;$i<@ALLOWED;$i++)
		{
			if(@ALLOWED[$i] eq @_[0])
			{
				return(1);
			}
		}
	}
	# Run through each entry in the denied
	# sites list and check each entry against
	# the currently requested site.
	else
	{
		for(my $i=0;$i<@DENIED;$i++)
		{
			if(@DENIED[$i] eq @_[0])
			{
				return(1);
			}
		}
	}

	return(0);
}

sub getAllowedDeniedSites($)
{
	my $PATH=shift;
	my @Stats;
	my $buffer;

	open(CONF,$PATH);
	@Stats=stat(CONF);
	read(CONF,$buffer,@Stats[7]);
	close(CONF);

	return(split(/\n/,$buffer));
}


########################## signal handling functions ###########################


sub debug
{
	$LOG=1;
	print STDOUT "Debugging mode enabled at: ".time()."\n\n";
}

sub reinit
{
	close($REMOTE);
	close($CLIENT);
	close($SERVER);

	$SERVER=undef;

	if($LOG==1){print STDOUT "Reinitializing Server at: ".time()."\n\n";}

	@ALLOWED=getAllowedDeniedSites($CONFPATH);
	@DENIED=getAllowedDeniedSites($DENYPATH);
	acceptConnections();
}

sub shutDown
{
	close($REMOTE);
	close($CLIENT);
	close($SERVER);

	exit();
}


########################### html filtering functions ###########################


sub filter($)
{
	my $doc=shift;
	my $sub;

	for(my $i=0;$i<@FILTERS;$i++)
	{
		$sub=@FILTERS[$i];
		$doc=&$sub($doc);
	}

	return($doc);
}

sub selectFilter($)
{
	my $doc=shift;

	$doc=~
		s{
			(<\s*select\s+)
			(.*?)
			(>)
		 }
		 {<select size="4" $2>}gixs;

	return($doc);
}

sub windowFilter($)
{
	my $doc=shift;

	if($address ne "localhost")
	{
		$doc=~
			s{(window.open)}
			 {(sore.open)}gixs;
	}

	return($doc);
}

sub targetFilter($)
{
	my $doc=shift;

	if($address ne "localhost")
	{
		$doc=~
			s{(target="(|_)(blank|top|window|new|main)")}
			 {}gixs;
	}

	return($doc);
}

sub templateFilter($)
{
	my $doc=shift;

	# Add filter code here

	return($doc);
}


############ command-line parsing and user-agent building functions ############


sub getVers
{
	my $version=`/bin/more /proc/version`;
	if($version=~/\b(Linux version)\s+(\S+)/i){$version=$2;}
	return($version);
}

sub verifyCommandLine
{
	foreach (@ARGV)
	{
		if(/(^-browser)(=)([0-1]{1})($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-denyfile)(=)([\w\/\.\s\"\']*)($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-filtering)(=)([0-1]{1})($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-hostfile)(=)([\w\/\.\s\"\']*)($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-log=)([0-1]{1})($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-maxcon)(=)([0-9]*)($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-port)(=)([0-9]*)($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-sitelock)(=)([0-1]{1})($)/i){$ARGS{lc($1)}=$3;}
		elsif(/(^-timeout)(=)([0-9]*)($)/i){$ARGS{lc($1)}=$3;}
		else{die("\nusage: qdp -[option]=[value]\n\n");}
	}
}


########################### error handling functions ###########################


sub forbidden
{
	my $return=<<THE_END;
<HTML>
  <HEAD>
    <TITLE>You can't go here.</TITLE>
  </HEAD>
  <BODY BGCOLOR="#FFFFFF">
    <BR><BR><BR><BR>
    <CENTER><H1>You can't go here.</H1></CENTER>
  </BODY>
<HTML>
THE_END

	print STDERR time(),"\nUnauthorized attempt to access site at: @_[0]\n";
	return($return);
}

sub unavailable
{
	my $return=<<THE_END;
<HTML>
  <HEAD>
    <TITLE>Temporarily unable to connect.</TITLE>
  </HEAD>
  <BODY BGCOLOR="#FFFFFF">
    <BR><BR><BR><BR>
    <CENTER><H1>Temporarily unable to connect.</H1></CENTER>
  </BODY>
<HTML>
THE_END

	print STDERR time(),"\nUnable to connect to site at: @_[0]\n";
	return($return);
}